# Rest Framework
from rest_framework import serializers

# django_register
# drf-serializer-prefetch
from django_register.rest_framework import RegisterField
from tests.models import City


class CitySerializer(serializers.ModelSerializer):
    country = RegisterField()

    class Meta:
        model = City
        fields = ("label", "country")



# Standard libraries
from typing import TYPE_CHECKING

# Django
from django.utils.translation import gettext_lazy as _

# Rest Framework
from rest_framework import serializers

# django_register
from django_register.base import Register

if TYPE_CHECKING:
    # Rest Framework
    from rest_framework.serializers import BaseSerializer


class RegisterField(serializers.CharField):
    def __init__(self, *args, **kwargs):
        self.register: Register = kwargs.pop("register", None)
        super().__init__(*args, **kwargs)

    def _get_register_from_parent(self, parent: "BaseSerializer"):
        assert hasattr(parent, "Meta") and hasattr(  # noqa: S101
            parent.Meta, "model"
        ), (
            "Cannot infer the register from the serializer field. "
            "Please provide a register or a Meta class with a model attribute."
        )

        field = parent.Meta.model._meta.get_field(self.field_name)

        try:
            register = field.register
        except AttributeError:
            raise ValueError(
                _(
                    "Cannot infer the register from the serializer field. "
                    "It is not a RegisterField."
                )
            )

        return register

    def bind(self, field_name: str, parent: "BaseSerializer") -> None:
        super().bind(field_name, parent)

        if self.register is None:
            self.register = self._get_register_from_parent(parent)

    def to_internal_value(self, data: str) -> str:
        return self.register.get_class(data)

    def to_representation(self, value: str) -> str:
        return self.register.get_key(value)



# Django
from django.test import TestCase

# Rest Framework
from rest_framework import serializers

# django_register
from django_register.rest_framework import RegisterField
from tests.models import City, CountryChoices


class CitySerialier(serializers.ModelSerializer):
    country = RegisterField()

    class Meta:
        model = City
        fields = ("label", "country")


class RegisterSerializerTestCase(TestCase):
    @classmethod
    def setUpTestData(cls) -> None:
        cls.paris = City.objects.create(label="Paris", country=CountryChoices.FRANCE)

    def test_serialization(self):
        serializer = CitySerialier(self.paris)
        self.assertEqual(serializer.data, {"label": "Paris", "country": "france"})

    def test_save(self):
        serializer = CitySerialier(
            data={"label": "Paris", "country": CountryChoices.FRANCE}
        )

        self.assertTrue(serializer.is_valid())
        city = serializer.save()

        self.assertEqual(city.country, CountryChoices.FRANCE)

    def test_wrong_field_type(self):
        class CitySerialier(serializers.ModelSerializer):
            label = RegisterField()
            country = RegisterField()

            class Meta:
                model = City
                fields = ("label", "country")

        serializer = CitySerialier(self.paris)

        with self.assertRaises(ValueError):
            serializer.data



# Django
from typing import Hashable
from django.core.exceptions import ValidationError
from django.db import models
from django.utils.deconstruct import deconstructible
from django.utils.translation import gettext_lazy as _


@deconstructible
class Register:
    def __init__(self):
        self._key_to_class = {}
        self._class_to_key = {}

        self._choices = []
        self._flatchoices = []

        self._fields = []

    def _get_class_lookup(self, klass):
        return klass if isinstance(klass, Hashable) else id(klass)

    def register(self, klass, db_key=None):
        if db_key is None:
            try:
                db_key = klass.label
            except AttributeError:
                raise ValueError(
                    _(
                        "The class {klass} does not have a label. Define "
                        "one or pass a db_key to be used as database value."
                    ).format(klass=klass)
                )

        if db_key in self._key_to_class:
            raise ValueError(_("Key {key} already registered.").format(key=db_key))

        if klass in self._class_to_key:
            raise ValueError(_("Class {klass} already registered.").format(klass=klass))

        self._key_to_class[db_key] = klass
        self._class_to_key[klass] = db_key

        self._choices.append((db_key, self._get_verbose_name(klass, db_key)))
        self._flatchoices.append((klass, self._get_verbose_name(klass, db_key)))

        for field in self._fields:
            if hasattr(field, "_choices"):
                field._choices = self._choices
            else:
                field.choices = self._choices

        return klass

    def add_field(self, field: models.Field):
        self._fields.append(field)

    def from_key(self, value):
        try:
            return self._key_to_class[value]
        except (KeyError, TypeError):
            raise ValidationError(
                _("Value {value} not a registered key.").format(value=value)
            )

    def from_class(self, value):
        try:
            class_lookup = self._get_class_lookup(value)
            return self._class_to_key[class_lookup]
        except KeyError:
            raise ValidationError(
                _("Value {value} not a registered class.").format(value=value)
            )

    def get_key(self, value):
        try:
            self.from_key(value)
        except ValidationError:
            return self.from_class(value)

        return value

    def get_class(self, value):
        try:
            self.from_class(value)
        except ValidationError:
            return self.from_key(value)

        return value

    @property
    def max_length(self):
        if self._key_to_class:
            return max(len(key) for key in self._key_to_class)

    @property
    def choices(self):
        return self._choices

    @property
    def flatchoices(self):
        return self._flatchoices

    def _get_verbose_name(self, klass, key):
        return getattr(klass, "verbose_name", key.replace("_", " ").title())

    def __iter__(self):
        return iter(self._key_to_class.values())


class RegisterList(list):
    register = None


class RegisterChoicesMeta(type):
    def __new__(mcs, name, bases, attrs):  # noqa: N804
        cls = super().__new__(mcs, name, bases, attrs)

        cls.register = Register()
        for label, member in cls._all_mapping.items():
            cls.register.register(member, db_key=label)

        return cls

    def _label_name(cls, name, obj):
        default_label = name.lower()
        return getattr(obj, "label", default_label)

    @property
    def _all_mapping(cls):
        return {
            cls._label_name(key, value): value
            for key, value in cls.__dict__.items()
            if not key.startswith("_") and key.isupper()
        }

    @property
    def choices(cls):
        choices = RegisterList()

        for label, obj in cls._all_mapping.items():
            default_verbose_name = label.replace("_", " ").title()
            verbose_name = getattr(obj, "verbose_name", default_verbose_name)

            choices.append((label, verbose_name))

        choices.register = cls.register
        return choices

    def __iter__(cls):
        return iter(cls.register)


class RegisterChoices(metaclass=RegisterChoicesMeta):
    def __new__(cls, klass):
        return cls.register.get_class(klass)


class RegisterField(models.CharField):
    description = _("Store a string, return the associated class")

    def __init__(self, *args, **kwargs):
        if "register" not in kwargs and "choices" not in kwargs:
            raise ValueError(_("You must provide choices to the RegisterField."))

        if "register" not in kwargs and not hasattr(kwargs["choices"], "register"):
            raise ValueError(_("Choices must be a RegisterChoices instance."))

        # When building the migrations, the register cannot be in the choices.
        # It will be passed individually, so we take it from there.
        self.register: Register = (
            kwargs.pop("register")
            if "register" in kwargs
            else kwargs["choices"].register
        )
        self.register.add_field(self)

        if "choices" not in kwargs:
            kwargs["choices"] = self.register.choices

        if "max_length" not in kwargs and (max_length := self.register.max_length):
            kwargs["max_length"] = max_length

        if "default" in kwargs:
            try:
                kwargs["default"] = self.register.get_key(kwargs["default"])
            except ValidationError:
                pass

        super().__init__(*args, **kwargs)

    def from_db_value(self, value, expression, connection):
        if not value:
            return value

        return self.register.get_class(value)

    def get_default(self):
        default = super().get_default()

        if default:
            return self.register.get_class(default)

        return default

    def to_python(self, value):
        if not value:
            return value

        return self.register.get_class(value)

    def get_prep_value(self, value):
        if not value:
            return value

        return self.register.get_key(value)

    def value_from_object(self, obj):
        value = super().value_from_object(obj)
        return self.get_prep_value(value)

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        kwargs.pop("choices", None)
        kwargs["register"] = self.register
        return name, path, args, kwargs

    def clean(self, value, model_instance):
        """
        We need to override clean because it runs the validations on the
        Python object instead of on the database string.
        """
        value = self.get_prep_value(value)
        self.validate(value, model_instance)
        self.run_validators(value)
        return self.to_python(value)

    def _get_flatchoices(self):
        return self.register.flatchoices

    flatchoices = property(_get_flatchoices)



# Django
from django.test import TestCase

# django_register
from django_register.base import Register
from tests.models import CountryChoices, CountryInfo


class RegisterTestCase(TestCase):
    @classmethod
    def setUpTestData(cls) -> None:
        cls.register: Register = CountryChoices.register

    def test_can_retrieve_obj(self):
        country_info = CountryInfo(1, capital="Max City")
        with self.assertRaises(ValueError):
            self.register.register(country_info)

        country_info.label = "max_country"
        self.register.register(country_info)

        self.assertEqual(self.register.get_class("max_country"), country_info)
        self.assertEqual(self.register.get_class(country_info), country_info)
        self.assertEqual(self.register.get_key(country_info), "max_country")
        self.assertEqual(self.register.get_key("max_country"), "max_country")

        country_info = CountryInfo(3_000_000_000, capital="Max Capital")
        self.register.register(country_info, db_key="max_big_country")

        self.assertEqual(self.register.get_class("max_big_country"), country_info)
        self.assertEqual(self.register.get_key(country_info), "max_big_country")

    def test_iter(self):
        for klass in self.register:
            self.assertIn(klass, CountryChoices)

    def test_cannot_register_twice(self):
        with self.assertRaises(ValueError):
            self.register.register(CountryChoices.CANADA, db_key="max_big_country")

        with self.assertRaises(ValueError):
            self.register.register(
                CountryInfo(population=2, capital="Some capital"),
                db_key="canada",
            )



# Django
from django.apps import AppConfig


class TestsConfig(AppConfig):
    name = "tests"

    def ready(self):
        from .models import cars_register, CarCompanies

        cars_register.register(CarCompanies("Toyota"), db_key="toyota")



# Django
from django.contrib import admin
from django.test import TestCase

# django_register
from tests.models import City, ContinentChoices, ContinentInfo


@admin.register(City)
class CityAdmin(admin.ModelAdmin):
    pass


class AdminTestCase(TestCase):
    @classmethod
    def setUpTestData(cls) -> None:
        cls.city = City.objects.create(label="Ottawa")

    def setUp(self):
        self.admin = CityAdmin(City, admin.site)

    def test_admin_choices(self):
        # choices was renamed to _choices in Django 5.0
        choices = getattr(
            self.admin.opts._forward_fields_map["available_food"],
            "choices",
        ) or getattr(
            self.admin.opts._forward_fields_map["available_food"], "_choices", None
        )
        self.assertEqual(
            choices,
            [("pizza", "Pizza")],
        )

    def test_admin_select(self):
        field = self.admin.opts._forward_fields_map["continent"]
        cleaned_value = field.clean("America", self.city)
        self.assertEqual(cleaned_value, ContinentChoices.AMERICA)

    def test_admin_flatchoices(self):
        field = self.admin.opts._forward_fields_map["continent"]
        self.assertEqual(
            field.flatchoices,
            [
                (ContinentChoices.AMERICA, "America"),
                (ContinentChoices.EUROPE, "Europe"),
            ],
        )

    def test_choices_gets_updated(self):
        field = self.admin.opts._forward_fields_map["continent"]
        attr = "choices" if hasattr(field, "choices") else "_choices"

        self.assertEqual(
            getattr(field, attr),
            [
                ("America", "America"),
                ("Europe", "Europe"),
            ],
        )

        register = ContinentChoices.register

        self.assertEqual(
            register.choices, [("America", "America"), ("Europe", "Europe")]
        )

        register.register(ContinentInfo(label="Asia"))

        field = self.admin.opts._forward_fields_map["continent"]

        self.assertEqual(
            register.choices,
            [("America", "America"), ("Europe", "Europe"), ("Asia", "Asia")],
        )
        self.assertEqual(
            getattr(field, attr),
            [
                ("America", "America"),
                ("Europe", "Europe"),
                ("Asia", "Asia"),
            ],
        )



# Django
from django.core.exceptions import ValidationError
from django.db import models
from django.test import TestCase

# django_register
from django_register.base import RegisterField
from tests.models import (
    CountryInfo,
    Neighborhood,
    cars_register,
    CarCompanies,
    City,
    CountryChoices,
)


class RegisterFieldTestCase(TestCase):
    @classmethod
    def setUpTestData(cls) -> None:
        cls.paris = City.objects.create(label="Paris", country=CountryChoices.FRANCE)
        cls.berlin = City.objects.create(label="Berlin", country=CountryChoices.GERMANY)

    def test_can_retrieve_obj(self):
        self.assertEqual(self.paris.country, CountryChoices.FRANCE)

    def test_field_errors(self):
        with self.assertRaises(
            ValueError, msg="You must provide choices to the RegisterField."
        ) as exc:
            RegisterField()
            self.assertEqual(
                exc.exception.args[0], "You must provide choices to the RegisterField."
            )

        with self.assertRaises(
            ValueError, msg="Choices must be a RegisterChoices instance."
        ):
            RegisterField(choices=["a", "b", "c"])

        RegisterField(register=CountryChoices.register)

    def test_filter(self):
        self.assertEqual(City.objects.filter(country=CountryChoices.FRANCE).count(), 1)
        self.assertEqual(City.objects.filter(country=CountryChoices.GERMANY).count(), 1)

        self.assertEqual(
            City.objects.filter(country=CountryChoices.FRANCE).first(), self.paris
        )
        self.assertEqual(
            City.objects.filter(country=CountryChoices.FRANCE).first().country,
            self.paris.country,
        )

    def test_default_value(self):
        city = City.objects.create(label="Ottawa")
        self.assertEqual(city.country, CountryChoices.UNITED_STATES)
        self.assertEqual(city._meta.get_field("country").default, "united_states")

    def test_set_wrong_default_value(self):
        City._meta.get_field("country").default = CountryInfo(12, capital="Max Capital")

        with self.assertRaises(ValidationError):
            City.objects.create(label="Ottawa")

    def test_fails_if_fetching_before_registering(self):
        with self.assertRaises(ValueError):
            cars_register.register(CarCompanies("Toyota"), db_key="toyota")

        hyundai_car = CarCompanies("Hyundai")

        with self.assertRaises(ValidationError):
            self.paris.car_companies = hyundai_car
            self.paris.save()

        cars_register.register(hyundai_car, db_key="hyundai")

        cars_register._class_to_key.pop(hyundai_car)
        cars_register._key_to_class.pop("hyundai")

    def test_changing_register_dynamically(self):
        with self.assertRaises(ValueError):
            cars_register.register(CarCompanies("Toyota"), db_key="toyota")

        hyundai_car = CarCompanies("Hyundai")
        cars_register.register(hyundai_car, db_key="hyundai")

        self.paris.car_companies = hyundai_car
        self.paris.save()
        self.assertEqual(self.paris.car_companies, hyundai_car)

        cars_register._class_to_key.pop(hyundai_car)
        cars_register._key_to_class.pop("hyundai")

    def test_annotations(self):
        Neighborhood.objects.create(label="Montparnasse", city=self.paris)

        neighborhood = Neighborhood.objects.annotate(
            country=models.F("city__country")
        ).first()

        self.assertNotEqual(
            neighborhood.country,
            "france",
        )
        self.assertEqual(
            neighborhood.country,
            CountryChoices.FRANCE,
        )



# Generated by Django 5.0a1 on 2023-10-25 14:11

import django.db.models.deletion
import django_register.base
from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="City",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("label", models.CharField(max_length=50)),
                (
                    "country",
                    django_register.base.RegisterField(
                        default="united_states",
                        max_length=13,
                        register=django_register.base.Register(),
                    ),
                ),
                (
                    "continent",
                    django_register.base.RegisterField(
                        blank=True,
                        max_length=7,
                        null=True,
                        register=django_register.base.Register(),
                    ),
                ),
                (
                    "available_food",
                    django_register.base.RegisterField(
                        blank=True,
                        max_length=5,
                        null=True,
                        register=django_register.base.Register(),
                    ),
                ),
                (
                    "car_companies",
                    django_register.base.RegisterField(
                        blank=True,
                        max_length=50,
                        null=True,
                        register=django_register.base.Register(),
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="Neighborhood",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("label", models.CharField(max_length=50)),
                (
                    "city",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE, to="tests.city"
                    ),
                ),
            ],
        ),
    ]



# Standard libraries
from pathlib import Path

# django_register
from setuptools import find_packages, setup

VERSION = "1.0.4"
DESCRIPTION = "A field that returns a python object."
this_directory = Path(__file__).parent
LONG_DESCRIPTION = (this_directory / "README.md").read_text()

setup(
    name="django-register-field",
    version=VERSION,
    author="Maxime Toussaint",
    author_email="m.toussaint@mail.com",
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    packages=find_packages(),
    install_requires=("django>=3.2.0",),
    url="https://github.com/MaxDude132/django-register-field",
)



# Standard libraries
from dataclasses import dataclass

# Django
from django.db import models

# django_register
from django_register import Register, RegisterChoices, RegisterField


@dataclass(unsafe_hash=True)
class CountryInfo:
    population: int
    capital: str


class CountryChoices(RegisterChoices):
    CANADA = CountryInfo(population=37_742_154, capital="Ottawa")
    FRANCE = CountryInfo(population=65_273_511, capital="Paris")
    GERMANY = CountryInfo(population=83_783_942, capital="Berlin")
    UNITED_STATES = CountryInfo(population=331_900_000, capital="Washington")


@dataclass(unsafe_hash=True)
class ContinentInfo:
    label: str


@dataclass(unsafe_hash=True)
class FoodInfo:
    verbose_name: str


food_register = Register()
food_register.register(FoodInfo("Pizza"), db_key="pizza")


@dataclass(unsafe_hash=True)
class CarCompanies:
    verbose_name: str


cars_register = Register()


class ContinentChoices(RegisterChoices):
    AMERICA = ContinentInfo(label="America")
    EUROPE = ContinentInfo(label="Europe")


class City(models.Model):
    label = models.CharField(max_length=50)
    country = RegisterField(
        choices=CountryChoices.choices, default=CountryChoices.UNITED_STATES
    )
    continent = RegisterField(choices=ContinentChoices.choices, null=True, blank=True)
    available_food = RegisterField(register=food_register, null=True, blank=True)
    car_companies = RegisterField(
        register=cars_register, null=True, blank=True, max_length=50
    )


class Neighborhood(models.Model):
    label = models.CharField(max_length=50)
    city = models.ForeignKey(City, on_delete=models.CASCADE)
